(******************************************************************************)
(*   #####   #####    ####      Formalization of the GPS program logic        *)
(*  ##       ##  ##  ##         by Turon, Vafeiadis and Dreyer                *)
(*  ##  ###  #####    ####                                                    *)
(*  ##   ##  ##          ##     Copyright (c) Viktor Vafeiadis, 2013          *)
(*   #####   ##       ####      See README.txt for license                    *)
(******************************************************************************)

Require Import Vbase Relations ClassicalDescription ClassicalChoice.
Require Import List Permutation Sorting extralib. 
Require Import c11 exps.
Require Import GpsModel GpsModelLemmas GpsGlobal GpsVisible.

Local Open Scope string_scope.

Set Implicit Arguments.

(******************************************************************************)
(** * 3.3.2. Rely *)
(******************************************************************************)

Lemma rely_step :
  forall b acts (NIN: ~ In b acts) amap sb rf mo L
    (PC : prefix_closed sb rf acts)
    (V : forall a (IN: In a acts), valid_node (b::acts) acts amap rf L a)
    (CC: compat amap sb rf L)
    (CO: conform (b::acts) amap mo L acts)
    (CONS: Consistent (b::acts) amap sb rf mo)
    (INrf: in_rf (b::acts) L b = res_emp)
    (INesc: in_esc (b::acts) L b = res_emp)
    (INsb: in_sb (b :: acts) L b <> Rundef)
    (INsbOK: forall a, ~ sb a b -> L (Tsb a b) = res_emp)
    (OUT : outgoing (b::acts) L b = res_emp)
    (ND : NoDup (b ::acts))
    (LR : label_restr (b::acts) amap sb rf L),
  exists L',
    << V : forall a (IN: In a acts), valid_node (b::acts) (b::acts) amap rf L' a >> /\
    << CC: compat amap sb rf L' >> /\
    << CO: conform (b::acts) amap mo L' acts >> /\
    << RELY: rely (in_sb (b::acts) L' b) (amap b) 
                  (in_sb (b::acts) L' b +!+ in_rf (b::acts) L' b) >> /\
    << INesc: in_esc (b::acts) L' b = res_emp >> /\
    << OUT : outgoing (b::acts) L' b = res_emp >> /\
    << Esb : forall a b, L' (Tsb a b) = L (Tsb a b) >> /\
    << EsbS : forall a, L' (TsbS a) = L (TsbS a) >> /\
    << LR : label_restr (b::acts) amap sb rf L' >>.
Proof.
  ins. 
  assert (PI: pres_in sb rf b acts). {
    split; ins.
      assert (X := SB); apply (proj1 CONS), proj1 in SB; ins; desf. 
      by exfalso; red in CONS; desf; edestruct ACYC; vauto.
    red in CONS; desc; red in FIN; desc.
    specialize (Crf b); desf; desf.
    exploit (CLOlab a); destruct (amap a); ins; desf;
      exfalso; edestruct ACYC; vauto.
  }

  unfold NW; case_eq (amap b); ins; desf.
   exists L; ins; rewrite INrf, res_plus_emp_r; repeat (split; eauto).
   by ins; apply valid_node_ext_helper; try rewrite H; auto.

Case "Aalloc". 
   exists L; ins; rewrite INrf, res_plus_emp_r; repeat (split; eauto).
   by ins; apply valid_node_ext_helper; try rewrite H; auto.

Case "Aload ATna".
   exists L; ins; rewrite INrf, res_plus_emp_r; repeat (split; eauto).
   by ins; apply valid_node_ext_helper; try rewrite H; auto.

   revert INsb; case_eq (in_sb (b :: acts) L b) as INsb; intros _.
   repeat eexists; ins.

   edestruct visible_na with (b:=b) (l:=l) (v:=v') as [b_rf ?]; eauto; desf.
     by unfold incoming; rewrite INsb, INrf, INesc, !res_plus_emp_r.
   assert (CONS':=CONS); red in CONS'; desc.
   specialize (Crf b); desf; desf.
   2: by edestruct Crf; eauto; instantiate; rewrite H.
   rewrite H in *; ins; desf.
   assert (X := Crf0); eapply is_write_weaken, OTH in X; ins.
   2: by apply PI in Heq. 
   rewrite mhbE in *; desf; ins; [by destruct (amap b_rf); ins; desf|].
   edestruct Cwr; try eapply Heq; try eapply X; try eapply Cmo; eauto.
   by rewrite H; destruct (amap b_rf); ins; desf. 

  rewrite mhbE; intro FHB; desf; try rewrite H in *; desf.
  pose proof (hb_in_acts2 CONS FHB); ins; desf.
    by edestruct ACYC; eauto.
  by destruct NIN; eauto using prefix_closed_hb. 
    
Case "Aload ATat".
  destruct (rf b) as [b_rf|] eqn: RF.
  2: exists L; ins; rewrite INrf, res_plus_emp_r; repeat (split; eauto).
     2: by ins; apply valid_node_ext_helper; try rewrite H; auto.
  2: exists res_emp; rewrite res_strip_emp, res_plus_emp_r; split; eauto.
  2: ins; exfalso.
  2: edestruct visible_atomic with (b:=b) (l:=l) as [b_rf' ?]; eauto; desf.
     2: eby unfold incoming; rewrite INrf, INesc, !res_plus_emp_r.
  2: red in CONS; desc.
  2: by specialize (Crf b); rewrite RF in *; eapply Crf; eauto; rewrite H.

  assert (b_rf <> b) by (intro; desf; red in CONS; desc; edestruct ACYC; vauto).
  assert (INrf': In b_rf acts) by (by apply PI in RF).

  destruct (classic (~ synchronizes_with amap rf b_rf b)) as [SW|SW]; [|apply NNPP in SW].
  {
    exists L; ins; rewrite INrf, res_plus_emp_r; repeat (split; eauto).
      by ins; apply valid_node_ext_helper; try rewrite H; auto.
    exists res_emp; rewrite res_strip_emp, res_plus_emp_r; split; eauto.
    ins; exfalso.
    edestruct visible_atomic with (b:=b) (l:=l) as [b_rf' ?]; eauto; desf.
      eby unfold incoming; rewrite INrf, INesc, !res_plus_emp_r.
    red in CONS; desc.
    destruct SW; red; rewrite H; repeat split; try done.
    destruct (classic (b_rf' = b_rf)); desf; [by destruct (amap b_rf); ins; desf|].
    red in Cmo; desc.
    specialize (Crf b); rewrite RF, H in *; ins; desf.
    edestruct (Cmo1 _ b_rf' H3 b_rf) as [MO|MO]; ins; eauto.
    2: by edestruct Cwr; eauto; instantiate;
          rewrite H; ins; destruct (amap b_rf').
    exploit (CO l0); try apply MO; eauto using pres_in_hb.
    rewrite H5; ins; desf; desf.
    specialize (V _ INrf'); red in V; desf.
    by destruct (amap b_rf); desf; ins; desf; desc; rewrite GUAR0 in *.
  }

  exists (upd L (Trf b_rf b) (res_strip (out_rf (b :: acts) L b_rf))); ins.
  unfold in_sb, in_esc, outgoing, out_sb, out_esc; ins.
  rewrite !map_upd_irr; try (by rewrite in_map_iff; intro; desf).
  unfold in_rf; ins.
  repeat first [rewrite upds | rewrite updo; [|by intro; desf]].

  destruct (In_NoDup_Permutation _ INrf') as (? & P & ?); ins; desc; [by inv ND|].

  do 8 try split; eauto. 
    ins; apply valid_node_ext_helper; try rewrite H; auto.
    ins; assert (NEQ: a <> b) by congruence.
    specialize (V _ IN). 
    unfold valid_node in *; ins; desf.
    eexists r, I; ins.
    rewrite !replace_out_rf; ins.
    unfold incoming, out_esc, out_sb, in_sb, in_rf, in_esc in *; ins.
    rewrite !map_upd_irr; try (by rewrite in_map_iff; intro; desf).
    unfold NW.
    rewrite !updo, escr_ffun_upd_irr; try (by intro X; desf); repeat split; ins.
    by rewrite updo; eauto; intro; desf; rewrite H in *. 
    eapply rsum_eq_empD in INrf; eauto. 
    by apply in_map, in_map_iff; eauto using in_cons.

  red; ins.
  destruct (classic (In (Trf b_rf b) E)) as [Z|]; [|by rewrite map_upd_irr; eauto].
  intro K.
  exploit (CC (undup (TrfS b_rf :: E))); try done; intros.
    by rewrite ?in_undup_iff in *; ins; desf; eauto.
  specialize (V _ INrf'); red in V; desf.
  rewrite <- VrfS', replace_edge_emp in K; try apply helper2; ins.
  eapply rsum_eq_empD in INrf; eauto.
  by apply in_map, in_map_iff; eauto using in_cons.

  red; ins; exploit CO; try eapply MO; eauto.
  by rewrite !replace_out_rf; ins; try congruence;
    try (by eapply (rsum_eq_empD _ INrf), in_map, in_map_iff; eauto using in_cons). 

  exists (out_rf (b :: acts) L b_rf); ins.
  change (L (Tsb b b) +!+ rsum (map L (map (Tsb^~ b) acts))) with (in_sb (b :: acts) L b).
  assert (DEF: in_sb (b :: acts) L b +!+ out_rf (b :: acts) L b_rf <> Rundef).
  {
    destruct (classic (exists c, rf c = Some b_rf /\ is_rmw (amap c) /\ In c acts)) as [N|N]; desf.
  {
    replace (in_sb (b :: acts) L b) with 
      (rsum (map L (map (Tsb^~ b) (filter (mydecf (sb^~ b)) acts)))). 
    intro; apply (CC (Trf b_rf c :: map (Tsb^~ b) (filter (mydecf (sb^~ b)) acts))).
        constructor; ins; [by rewrite in_map_iff; intro; desf|].
        by eapply nodup_map; ins; try solve [by apply nodup_filter; inv ND|congruence]. 
       (*by ins; desf; rewrite in_map_iff in *; desf; rewrite filter_In in *; desf.*)
      ins; desf; try rewrite in_map_iff in *; rewrite mhbE in *; desf; ins; desf;
      try rewrite filter_In in *; desf;
      try by red in CONS; desc; destruct (ACYC b_rf); eauto using clos_trans.
    by unfold mydecf in *; desf;
       eapply no_rf_rmw_overtaken with (a:=x0) (b:=b); eauto using hb_trans.
    by unfold mydecf in *; desf; 
       eapply no_rf_rmw_overtaken with (a:=b0) (b:=b); eauto using hb_trans.
    by unfold mydecf in *; desf; red in CONS; desc; 
       destruct (ACYC b0); eauto using clos_trans.
    by simpl; specialize (V _ INrf'); red in V; desf; rewrite res_plusC, VrfM. 
    unfold in_sb; ins; rewrite INsbOK, res_plus_emp_l, rsum_map_map_filter; auto.
    by intro; red in CONS; desf; edestruct ACYC; vauto.
  }
  { intro; apply (CC (TrfS b_rf :: map (Tsb^~ b) acts)).
        constructor; ins; [by rewrite in_map_iff; intro; desf|].
        eapply nodup_map; ins; try solve [by inv ND|congruence]. 
      ins; desf; rewrite in_map_iff, mhbE in *; desf; ins; desf.
      by red in CONS; desc; destruct (ACYC b_rf); eauto using clos_trans.
     by apply NIN; clear - PC IN0 HB; red in PC; desf; induction HB as [? ? [|[]]|]; eauto.
    ins.
    rewrite res_plusC in H2; revert H2; unfold in_sb; ins.
    unfold outgoing, out_sb in OUT; rewrite !res_plusA, res_plusAC in OUT.
    rewrite (rsum_eq_empD _ (proj1 (res_plus_eq_empD _ _ OUT)) (L (Tsb b b))), res_plus_emp_l in *; vauto.
    by specialize (V _ INrf'); red in V; desf; rewrite <- VrfS in H2; ins; eauto.
  }
  }

  ins; desf.
  split; ins.

    edestruct visible_atomic with (b:=b) (l:=l) as [b_rf' ?]; eauto; desf.
      by unfold incoming; rewrite INrf, INesc, !res_plus_emp_r; eauto.
    red; unfold NW.
    destruct (classic (b_rf' = b_rf)); desf.

   rewrite H2, H6; repeat (eexists; try edone).

   specialize (V _ INrf'); red in V; desf. 
   assert (is_writeLV (amap b_rf) l v).
     clear - RF CONS H; red in CONS; desc.
     by specialize (Crf b); rewrite RF, H in *; ins; desf.
   by destruct (amap b_rf); ins; desf; 
     desc; (try by rewrite GUAR0 in *; ins);
     eauto using aguar_interp_prot.

    red in CONS; desc; red in Cmo; desc.
    specialize (Crf b); rewrite RF, H in *; ins; desf.
    edestruct (Cmo1 _ b_rf' H4 b_rf) as [MO|MO]; ins; eauto.
    2: by edestruct Cwr; eauto; instantiate;
          rewrite H; ins; destruct (amap b_rf').

    exploit (CO l0); try apply MO; eauto using pres_in_hb.

    rewrite H2, H6; ins.
    destruct (out_rf (b :: acts) L b_rf) eqn: Y; ins; destruct (pmap l0) eqn: Y'; ins; desf.
    repeat (eexists; eauto using slist_lt_trans).
    specialize (V _ INrf'); red in V; desf.
    destruct (amap b_rf); ins; desf; rewrite Y in *; 
       desc; eauto using aguar_interp_prot.
    by unfold res_emp in *; desf.

  rewrite !(Permutation_rsum (Permutation_map _ (Permutation_map (Trf^~ b) P))); ins.
  rewrite upds, !map_upd_irr; try (by rewrite in_map_iff; intro; desf).
  revert INrf; unfold in_rf; simpl.
  rewrite !(Permutation_rsum (Permutation_map _ (Permutation_map (Trf^~ b) P))); ins;
     apply res_plus_eq_empD in INrf; desf; apply res_plus_eq_empD in INrf0; desf.
     rewrite INrf, INrf1, res_plus_emp_l, res_plus_emp_r; split; ins.
  intro M; destruct DEF.
  by rewrite <- (res_plus_strip_idemp (out_rf _ _ _)), res_plusAC, M, res_plusC.

  by unfold out_rf; rewrite map_upd_irr, updo; ins; rewrite in_map_iff; intro; desf. 

  by ins; rewrite updo.
  by ins; rewrite updo.

Case "Astore ATna".
   exists L; ins; rewrite INrf, res_plus_emp_r; repeat (split; eauto).
   by ins; apply valid_node_ext_helper; try rewrite H; auto.

Case "Astore ATat".
  exists L; ins; rewrite INrf, res_plus_emp_r; repeat (split; eauto).
    by ins; apply valid_node_ext_helper; try rewrite H; auto.
  ins.
  exploit exists_immediate_helper; eauto.
     eby unfold incoming; rewrite INrf, INesc, !res_plus_emp_r.
    by rewrite H.
  intro Z; desf.

  desf; generalize (V _ INa); intro V'; red in V'; desf.
  assert (NEQ: a <> b) by congruence.
  assert (DEF: in_sb (b :: acts) L b +!+ L (TrfS a) <> Rundef).
    by eapply helper_plus_insb_rfS_def; eauto.

  assert (K := res_plus_get1 _ _ l DEF).
  unfold env_move, NW.

  assert (AG: exists v0, atomic_guar r l v0 (out_sb (b :: acts) L a)
                                            (out_rf (b :: acts) L a)).
    by red in GUAR; desf; desf; ins; desf; vauto.

  desf; generalize AG; red in AG; desc; intros; clear AG1.
  eexists , (out_rf (b :: acts) L a); revert DEF K; rewrite H0, Grf in *;
  rewrite VrfS, Grf; try done; eauto using no_rmw_between.
  ins; desf; eexists,; repeat split; eauto using aguar_interp_prot. 

  edestruct visible_atomic with (b:=b) (l:=l) as [c ?]; eauto; desf.
    eby unfold incoming; rewrite INrf, INesc, !res_plus_emp_r.
  destruct (eqP c a); [by subst; rewrite Grf in *; desf|].
  assert (In c acts).
    red in CONS; desf; ins; exploit (proj1 FIN c); ins; desf; 
    [by destruct (amap c)|by exfalso; eauto].
  assert (MOca: mo c a).
    red in CONS; desc; red in Cmo; desc.
    edestruct (Cmo1 l); try eapply n; ins.
    exfalso; eapply MO; try edone; eapply Cmo3; eauto.
    by rewrite H.
  eapply (CO l) in MOca; eauto. 
  by red in MOca; desf; desf; eauto using slist_lt_trans.

Case "Armw".
   destruct (rf b) as [b_rf|] eqn: RF.
   2: exists L; ins; rewrite INrf, res_plus_emp_r; repeat (split; eauto).
     2: by ins; apply valid_node_ext_helper2; try congruence; auto.
   2: exists res_emp; rewrite res_plus_emp_r; split; eauto.
   2: ins; exfalso.
   2: edestruct visible_atomic with (b:=b) (l:=l) as [b_rf' ?]; eauto; desf.
      2: eby unfold incoming; rewrite INrf, INesc, !res_plus_emp_r.
   2: red in CONS; desc.
   2: by specialize (Crf b); rewrite RF in *; eapply Crf; eauto; rewrite H.

  assert (b_rf <> b) by (intro; desf; red in CONS; desc; edestruct ACYC; vauto).
  assert (INrf': In b_rf acts) by (by apply PI in RF).

  destruct (classic (~ synchronizes_with amap rf b_rf b)) as [SW|SW]; [|apply NNPP in SW].
  {
    exists L; ins; rewrite INrf, res_plus_emp_r; repeat (split; eauto).
      ins.
      destruct (eqP a b_rf); desf; 
       [|by apply valid_node_ext_helper2; try congruence; auto].
      specialize (V _ IN); red in V; red; unfold NW; desf.
      exists r, I; repeat (split; try done); ins; desf; eauto.
      unfold synchronizes_with in SW.
      rewrite H in *; ins.
      red in CONS; desc.
      specialize (Crf c); rewrite RF in *; desf.
      destruct (amap b_rf); ins; desf; desc; try by destruct SW. 
      by rewrite GUAR0; apply (rsum_eq_empD _ INrf), in_map, in_map_iff; eauto using in_cons.
    exists res_emp; rewrite res_plus_emp_r; split; eauto.
    ins; exfalso.
    edestruct visible_atomic with (b:=b) (l:=l) as [b_rf' ?]; eauto; desf.
      eby unfold incoming; rewrite INrf, INesc, !res_plus_emp_r.
    red in CONS; desc.
    destruct SW; red; rewrite H; repeat split; try done.
    destruct (classic (b_rf' = b_rf)); desf; [by destruct (amap b_rf); ins; desf|].
    red in Cmo; desc.
    specialize (Crf b); rewrite RF, H in *; ins; desf.
    edestruct (Cmo1 _ b_rf' H3 b_rf) as [MO|MO]; ins; eauto.
    2: by edestruct Cwr; eauto; instantiate;
          rewrite H; ins; destruct (amap b_rf').
    exploit (CO l0); try apply MO; eauto using pres_in_hb.
    rewrite H5; ins; desf; desf.
    specialize (V _ INrf'); red in V; desf.
    by destruct (amap b_rf); desf; ins; desf; desc; rewrite GUAR0 in *.
  }

   assert (DEF: in_sb (b :: acts) L b +!+ L (TrfS b_rf) <> Rundef). {
     replace (in_sb (b :: acts) L b) with 
       (rsum (map L (map (Tsb^~ b) (filter (mydecf (sb^~ b)) acts)))). 
     unfold in_sb; rewrite res_plusC.
     intro; eapply (CC (TrfS b_rf :: map (Tsb^~ b) (filter (mydecf (sb^~ b)) acts))).
        constructor; [by rewrite in_map_iff; intro; desf|].
        by eapply nodup_map; try eapply nodup_filter; ins; try solve [by inv ND|congruence]. 
      by ins; desf; ins; try rewrite in_map_iff, mhbE in *; desf; ins; desf; 
         try rewrite filter_In in *; desf; eauto using prefix_closed_hb.
      by ins; rewrite in_map_iff; intro; desf.
     unfold in_sb; ins; rewrite INsbOK, res_plus_emp_l, rsum_map_map_filter; auto.
     by intro; red in CONS; desc; edestruct ACYC; vauto.
   }

  assert (LrfS: L (TrfS b_rf) = out_rf (b :: acts) L b_rf).
    specialize (V _ INrf'); red in V; desf.
    eapply VrfS; ins. 
    assert (b = c) by (by eapply no_rf_rmw_double; eauto; instantiate; try rewrite H). 
    desf.

  exists (upd (upd L (TrfS b_rf) (res_strip (out_rf (b :: acts) L b_rf)))
                     (Trf b_rf b) (out_rf (b :: acts) L b_rf)).
  unfold outgoing, incoming, in_sb, in_esc, out_sb, out_esc; rupd.
  rewrite !map_upd_irr; try (by rewrite in_map_iff; intro; desf).
  rewrite replace_out_rf2; ins; vauto.
  rewrite in_rf_upds; ins; eauto.
  2: by unfold in_rf; rewrite !map_upd_irr; 
        try (by rewrite in_map_iff; intro; desf). 

  split.
    intros; specialize (V _ IN); red in V; desf.
    exists r, I; unfold NW.
    rewrite !escr_ffun_upd_irr; ins.
    rewrite replace_out_rf2; ins; vauto.
    unfold outgoing, incoming, in_sb, in_rf, in_esc, out_sb, out_esc.
    rewrite !map_upd_irr; try (by rewrite in_map_iff; intro; desf).

    repeat (split; auto; try by rupd). 
      ins; desf; rupd; eauto; congruence.
      ins; desf; rupd; eauto; intro; desf; eauto.
      by eapply (H1 b); eauto; rewrite H.
    destruct (eqP a b_rf); desf; rupd; auto using res_strip_idemp; congruence.
  split.
    apply compat_upd2; ins; rewrite LrfS, <- ?res_plusA, ?res_plus_strip_l; vauto.
    by eexists; apply res_plus_strip_l.
  split.
    by red; ins; rewrite !replace_out_rf2; ins; eauto.
  split.
    exists (out_rf (b :: acts) L b_rf),; ins; repeat (split; ins).

    assert (is_writeLV (amap b_rf) l v).
      clear - H CONS RF; red in CONS; desc.
      by specialize (Crf b); rewrite RF in Crf; desf; rewrite H in *; ins; desf.

    rewrite LrfS in *.

    edestruct visible_atomic with (b:=b) (l:=l) as [b_rf' ?]; eauto; desf.
      eby unfold incoming; rewrite INrf, INesc, !res_plus_emp_r.

    specialize (V _ INrf'); red in V; desf.
    generalize (proj1 (proj2 SW)).
    red; unfold NW.

    destruct (eqP b_rf' b_rf); desf.

    rewrite H1, H6.

  repeat (eexists; ins).
  generalize (proj1 (proj2 SW)).
  destruct (amap b_rf); ins; desf; desc; eauto using aguar_interp_prot. 

  red in CONS; desc; red in Cmo; desc.
  eapply Cmo1 in n; eauto; desf. 

  eapply CO in n; eauto using pres_in_hb.
  rewrite H6 in *; ins; desf; desf.
  rewrite H1.

  eexists,; repeat (split; ins); eauto using slist_lt_trans.
  destruct (amap b_rf); ins; desf; desc; eauto using aguar_interp_prot. 

  eapply (Cwr _ _ H3) in n; ins.
  by destruct (amap b_rf'); try done; destruct (amap b); ins; desf.

  by rewrite <- LrfS.

  repeat (split; eauto using in_cons); ins; rupd.
Qed.


Lemma rely_compat_post :
  forall b acts amap sb rf 
    (LAST: forall x, ~ happens_before amap sb rf b x) L
    (LR  : label_restr (b::acts) amap sb rf L)
    (CC: compat amap sb rf L)
    (ND : NoDup (b::acts))
    (OUT : outgoing (b::acts) L b = res_emp),
  compat amap sb rf 
    (upd L (TescS b) 
         (L (TescS b) +!+ in_sb (b::acts) L b +!+ in_rf (b::acts) L b)).
Proof.
  red; ins; destruct (classic (In (TescS b) E)).
  2: by rewrite map_upd_irr; eauto.
  eapply In_NoDup_Permutation in H; desf.
  rewrite !(Permutation_rsum (Permutation_map _ H)); simpl; rupd.
  rewrite !map_upd_irr; try done.
  assert (NoDup l') by (eapply Permutation_nodup in H; eauto; inv H).
  intro; apply (CC (map (Tsb^~ b) (filter (mydecf (sb^~ b)) acts) ++ 
                    map (Trf^~ b) (filter (mydecf (synchronizes_with amap rf ^~ b)) acts) ++ 
                    filter (mydecf (fun te => (te_src te) <> b /\ te_tgt te <> Some b)) l')).
   repeat rewrite nodup_app; repeat split; unfold disjoint; try done; 
     try apply nodup_map; try apply nodup_filter; ins; try congruence;
     rewrite ?in_app_iff, ?in_map_iff, ?filter_In in *; unfold mydecf in *; 
     desf; ins; desf; try tauto;
       try by inv ND.

   inv ND.
  intros; rewrite !in_app_iff, !in_map_iff, !filter_In in *; desf; ins; desf; eauto;
    rewrite ?filter_In in *; desf; unfold mydecf in *; desf; desf;
    try (eby edestruct LAST; eauto);
    try (solve [rewrite mhbE in *; desf; eby edestruct LAST; eauto]);
    try (rewrite mhbE in *; desf; try (eby edestruct LAST; eauto); []);
    try (solve [match goal with H : Permutation E (?x :: _) |- _ =>
                  eapply IND with (te := x); try edone; 
                  eauto using Permutation_in, Permutation_sym, in_eq, in_cons; 
                  eapply mhb_trans; vauto
                end]).
  eapply IND in HB; try edone; 
    eauto using Permutation_in, Permutation_sym, in_eq, in_cons. 


  inv ND; rewrite !map_app, !rsum_app.
  apply res_plus_eq_empD in OUT; desc.
  apply res_plus_eq_empD in OUT0; desc.
  apply res_plus_eq_empD in OUT1; desc.
  apply res_plus_eq_empD in OUT1; desc; rewrite OUT1, res_plus_emp_l in *.
  apply res_plus_eq_empD in OUT; desc.
  apply res_plus_eq_empD in OUT0; desc.
  rewrite <- H2, res_plusA; unfold in_sb, in_rf; ins.
  rewrite (proj1 (res_plus_eq_empD _ _ OUT4)), res_plus_emp_l.
  rewrite (proj1 (res_plus_eq_empD _ _ OUT5)), res_plus_emp_l.
  f_equal; [|f_equal]; try apply rsum_map_map_filter; 
    try apply rsum_map_filter; ins; try eby apply LR.
  apply not_and_or in H4; desf.
    by apply NNPP; intro M; destruct H4; intro; desf; destruct M, x; ins;
       apply LR; ins; intro; edestruct LAST; eauto.
  destruct x; apply LR; ins; intro; desf; destruct H4; vauto;
    try intro; desf; try (by edestruct LAST; eauto);
    try match goal with 
      H : Permutation _ (?x :: _), H' : In ?y _ |- _ =>
        by eapply IND with (te := x) (te' := y); try edone;
          eauto using Permutation_in, Permutation_sym, in_eq, in_cons; vauto
    end.
Qed.



(******************************************************************************)
(** * 3.3.3. Ghost moves *)
(******************************************************************************)

Lemma incoming_upd_ext_esc :
  forall acts (ND: NoDup acts) a (IN: In a acts) L b rr,
    incoming acts (upd L (Tesc a b) (rr +!+ L (Tesc a b))) b =
    rr +!+ incoming acts L b.
Proof.
  ins; rewrite !(res_plusC rr); unfold incoming, in_sb, in_rf.
  rewrite !map_upd_irr; try (by rewrite in_map_iff; intro; desf).
  rewrite !res_plusA; do 2 f_equal.
  unfold in_esc.
  eapply In_NoDup_Permutation in IN; desf.
  rewrite !(Permutation_rsum (Permutation_map _ (Permutation_map _ IN))); simpl; rewrite upds.
  rewrite map_upd_irr, !res_plusA; f_equal; eauto using res_plusC.
  rewrite in_map_iff; intro; desf.
Qed.

Lemma label_restr_fin acts amap sb rf mo L :
  Consistent acts amap sb rf mo ->
  label_restr acts amap sb rf L ->
  exists tel, forall te (NIN: ~ In te tel), L te = res_emp.
Proof.
  ins.
  exists 
    (flatten (map (fun a => map (Tsb a) acts) acts) ++
     flatten (map (fun a => map (Trf a) acts) acts) ++
     flatten (map (fun a => map (Tesc a) acts) acts) ++
     map TsbS acts ++ map TrfS acts ++ map TescS acts ++ map TcondS acts).
  ins.
  rewrite !in_app_iff, !in_flatten_iff, !in_map_iff in *.
  repeat match goal with 
           H : ~ (_ \/ _) |- _ => apply not_or_and in H; desc 
         end.
  destruct te; ins; apply H0; intro; eauto;
    exploit Consistent_hb_in; eauto; intro; desf; eauto 8 using in_map.
Qed.

Lemma res_gget_max_list l :
  exists n, forall n' (LT: n < n') r , In r l -> res_gget r n' = elem_gres_emp.
Proof.
  induction l; ins; desf; [by exists 0|].
  destruct a as [|p [f [n' C]] e].
    by exists n; ins; desf; eauto.  
  exists (n + n'); ins; desf; ins; eauto with arith.
Qed.

Lemma orR r r' : r || r' <-> r \/ r'.
Proof. case orP; split; vauto. Qed.

Lemma ghost_step :
  forall b acts (NIN: ~ In b acts) amap sb rf mo r P
    (CONS: Consistent (b::acts) amap sb rf mo)
    (ND: NoDup (b :: acts))
    (PC : prefix_closed sb rf acts)
    (PI: pres_in sb rf b acts)
    (GM: ghost_move r P) L
    (V : forall a (IN: In a acts), valid_node (b::acts) (b::acts) amap rf L a)
    (CC: compat amap sb rf (upd L (TescS b) (L (TescS b) +!+ r)))
    (CO: conform (b::acts) amap mo L acts) I 
    (VG: valid_ghost (incoming (b::acts) L b)
                     (r +!+ out_esc (b::acts) L b +!+ L (TcondS b)) I) 
    (VI : forall sigma r (GIN: In (sigma, r) I), 
            exists P P', interp_escr sigma = (P, P') /\ P' r) 
    (RS: res_lt (res_strip (incoming (b::acts) L b)) r)
    (VES: res_lt (rsum (map snd (filter (escr_ffun L (b::acts)) I))) (L (TescS b)))
    (EO: forall c, L (Tesc b c) = res_emp)
    (LR  : label_restr (b::acts) amap sb rf L),
  exists L' I r,
    << IN': P r >> /\
    << V : forall a (IN: In a acts), valid_node (b::acts) (b::acts) amap rf L' a >> /\
    << CC: compat amap sb rf (upd L' (TescS b) (L' (TescS b) +!+ r)) >> /\
    << CO: conform (b::acts) amap mo L' acts >> /\
    << VG: valid_ghost (incoming (b::acts) L' b)
                       (r +!+ out_esc (b::acts) L' b +!+ L' (TcondS b)) I >> /\ 
    << RS: res_lt (res_strip (incoming (b::acts) L' b)) r >> /\
    << VI  : forall sigma r (GIN: In (sigma, r) I), 
            exists P P', interp_escr sigma = (P, P') /\ P' r >> /\
    << VES: res_lt (rsum (map snd (filter (escr_ffun L' (b::acts)) I))) (L' (TescS b)) >> /\
    << EO: forall c, L' (Tesc b c) = res_emp >> /\
    << Esb: forall a b, L' (Tsb a b) = L (Tsb a b) >> /\
    << Erf: forall a b, L' (Trf a b) = L (Trf a b) >> /\
    << EsbS: forall a, L' (TsbS a) = L (TsbS a) >> /\
    << ErfS: forall a, L' (TrfS a) = L (TrfS a) >> /\
    << LR  : label_restr (b::acts) amap sb rf L' >>.
Proof.
  induction 6; ins; unfold NW; desf.
    by eexists L, I, r; repeat (split; ins).

Case "transitivity".
  exploit IHGM; eauto.
  clear r V CC CO VG EO RS IHGM GM GM' VI VES LR.
  intro; desf; exploit H; eauto.
  by intro; desf; repeat (eexists; try edone); eauto using eq_trans.

Case "Ghost allocation".
  assert (MM: exists mm, 
                res_gget r mm = elem_gres_emp /\ 
                forall te, res_gget (L te) mm = elem_gres_emp).
    clear - CONS LR.
    destruct (label_restr_fin CONS LR) as [myL EMP].
    destruct (res_gget_max_list (r :: map L myL)) as [mm X].
    exists (S mm),; ins. eapply X; ins; vauto.
    destruct (classic (In te myL)); eauto using in_map.
    by rewrite EMP; ins.

  desf; assert (DEF: r <> Rundef) by (intro; subst; red in VG; desf).

  eexists L, I, (res_plus r (Rdef (fun _ => p_bot) (gres_one mm gone) (fun _ => false))),; 
    try edone; repeat (split; try done).
    by eapply PEQ; eauto using gres_plus_emp_r, res_gget_def. 
  red; red; ins; eapply CC; eauto.

    revert H.
    destruct (classic (In (TescS b) E)) as [C|C]; 
      [eapply In_NoDup_Permutation in C; desf;
       rewrite !(Permutation_rsum (Permutation_map _ C)); simpl; rewrite !upds|];
    rewrite !map_upd_irr; try done.
    rewrite !res_plusA, !(res_plusAC (Rdef _ _ _)), res_plusC; intro.
    apply NNPP; intro.
     revert H; apply res_gget_def; ins.
    unfold out_esc; repeat (rewrite res_gget_plus; right; split; ins).
    by rewrite res_gget_rsum; right; ins; rewrite in_map_iff in *; desf.

    red in VG; ins; desf.
    rewrite !res_plusA, res_plusAC, res_plusC.
    eapply res_gget_def2 with (mm := mm) (gone := gone) in VG0; eauto; desf.
    repeat (eexists; eauto).
    unfold out_esc; repeat (rewrite res_gget_plus; right; split; ins).
    by rewrite res_gget_rsum; right; ins; rewrite in_map_iff in *; desf.

  by eapply res_lt_trans, res_lt_plus_r, res_lt_refl.

Case "Ghost update".
  assert (GG: exists g'', gres_plus (gres_strip g) g' = Some g''). { 
    generalize (proj2 (GDEF (gres_strip g))); rewrite gres_plus_strip_idemp, gres_plusC. 
    destruct (gres_plus (gres_strip g) g'); vauto.
    by intro X; specialize (X eq_refl).
  }
  desc.
  assert (GDEF' : forall gF,
         gres_plus g gF = None <-> gres_plus g'' gF = None).
    clear - GG GDEF; split; ins; rewrite GDEF in *.
      by case_eq (gres_plus g'' gF) as EQ; destruct (gres_plusA GG EQ); desf.
    rewrite gres_plusC in GG.
    apply GDEF; apply (gres_plusA2 GG) in H; desf; [|apply GDEF in H]; 
    case_eq (gres_plus g gF) as EQ;
    by destruct (gres_plusA (gres_plus_strip_idemp _) EQ); desf.

  eexists L, I, (Rdef p g'' e),; try edone; repeat (split; try done).
    assert (XX: Rdef p g'' e = 
          Rdef p g' e +!+ Rdef (fun _ => p_bot) (gres_strip g) (fun _ => false)).
      by rewrite res_plusC; ins; rewrite lift_plus_emp_l, GG.
    by rewrite XX; eapply RP_ext; congruence. 

    repeat red; intros; eapply CC; eauto; revert H.
    destruct (classic (In (TescS b) E)) as [C|C]; 
      [eapply In_NoDup_Permutation in C; desf;
       rewrite !(Permutation_rsum (Permutation_map _ C)); simpl; rewrite !upds|];
    rewrite !map_upd_irr; try done.
    rewrite !res_plusA, 2!(res_plusAC (Rdef _ _ _)); simpl; desf.
    rewrite <- GDEF' in *; congruence.
  clear - VG GDEF'; unfold valid_ghost in *; ins; desf.
  2: by rewrite <- GDEF' in *; congruence.
  by rewrite VG; do 7 eexists; eauto. 

  rewrite res_lt_strip_l in *.
  eapply res_lt_trans; eauto.
  exists (Rdef (fun _ => p_bot) (gres_strip g') (fun _ => false)).
  clear - GG; ins.
  apply gres_strip_plus in GG; rewrite gres_strip_idemp in GG; rewrite GG.
  rewrite lift_plusC, lift_plus_emp_l; ins; auto using prot_plusC; f_equal; ins.
  by extensionality l; rewrite orbF.

Case "Escrow introduction". 
  assert ((r1 +!+ Rdef (fun _ : nat => p_bot) gres_emp (eq_op^~ sigma)) +!+
          res_strip r2 <> Rundef).
    by rewrite res_plusC, res_plusAC, <- res_plusA; auto with rpd.
  eexists (upd L (TescS b) (L (TescS b) +!+ r2)), ((sigma, r2) :: I),
          ((r1 +!+ Rdef (fun _ : nat => p_bot) gres_emp (eq_op^~ sigma)) +!+ res_strip r2),; 
    try edone. 
    by eapply RP_ext; eauto.
  repeat (split; auto using in_eq, in_cons; try done); try (by intros; apply updo). 
  clear - NIN V; ins; specialize (V _ IN); unfold valid_node in *; desf.
  eexists r, I; ins.
  unfold incoming, out_esc, out_sb, out_rf, in_sb, in_rf, in_esc in *; ins.
  rewrite !map_upd_irr; try (by rewrite in_map_iff; intro; desf).
  unfold NW.
  repeat split; try by unfold upd; ins; desf; ins; eauto.
  rewrite escr_ffun_upd_irr, updo; ins; congruence.

  rewrite updss, upds.
  replace ((L (TescS b) +!+ r2) +!+
           (r1 +!+ Rdef (fun _ : nat => p_bot) gres_emp (eq_op^~ sigma)) +!+
           res_strip r2)
  with ((L (TescS b) +!+ r2) +!+
           (r1 +!+ Rdef (fun _ : nat => p_bot) gres_emp (eq_op^~ sigma))).
  2: rewrite !res_plusA; f_equal; rewrite !(res_plusC r2), !res_plusA.
  2: by rewrite (res_plusC (res_strip r2)), res_plus_strip_idemp. 
  clear - NIN CC; repeat red; ins; eapply CC; eauto. 
  revert H. 
  destruct (classic (In (TescS b) E)) as [C|C]; 
    [eapply In_NoDup_Permutation in C; desf;
     rewrite !(Permutation_rsum (Permutation_map _ C)); simpl; rewrite !upds|];
  rewrite !map_upd_irr; eauto. 
  by rewrite !res_plusA, (res_plusAC r1), !(res_plusAC (Rdef _ _ _)); simpl; desf;
     rewrite ?gres_plusI, ?lift_plus_emp_l in *.
   
  clear - NIN CO; red; ins.
  unfold outgoing, out_sb, out_esc, out_rf; ins. 
  rewrite !map_upd_irr; try (by rewrite in_map_iff; intro; desf).
  unfold upd; desf; eapply CO; ins.

  unfold incoming, out_esc, in_esc, in_sb, in_rf in *; ins.
  rewrite !map_upd_irr; try (by rewrite in_map_iff; intro; desf).
  unfold upd; desf.
  rewrite !res_plusA in *.
  rewrite (res_plusAC _ r1), 2!(res_plusAC r2).
  rewrite <- (res_plusA r2), res_plus_strip_idemp.
  red in VG; desf; rewrite VG, VG0; ins; rewrite lift_plus_emp_l, gres_plusI; ins. 
  do 8 eexists; ins; desf; eauto; try apply/orP; try rewrite VG1 in *; desf; 
    eauto using eqxx.

  clear - H RS RDEF.
  rewrite incoming_upd_irr; ins.
  rewrite res_lt_strip_l, !res_strip_plus, !res_strip_idemp in *; ins.
    by rewrite res_plusA, res_plusAC; auto using res_lt_plus_l.
  by intro X; rewrite X in *.
  
  by ins; desf; eauto.

  by rewrite escr_ffun_upd_irr, upds, res_plusC; ins; desf; ins; 
     auto using res_lt_frame_l, res_lt_plus_l.

  by ins; rupd.

Case "Escrow elimination".
  destruct (classic (res_escr (incoming (b :: acts) L b) sigma)).

  SCase "escrow introduced at node a <> b".

    exploit visible_escrow; eauto; intros (a & NN); desc.
    assert (INa : In a acts) by eauto using pres_in_hb.
    assert (NEQ: a <> b) by congruence.
    exploit (V a); eauto. 
    intros (rA & IA & VA); desc.
     unfold outgoing in NN1; exploit guar_escr; eauto.
     by intro X; rewrite X in *; red in VG0; desc.
    rewrite (vghost_escr VG0); intro; desf.
    assert (ERES: P2 rr) by (exploit VI0; eauto; instantiate; rewrite ESCR; ins; desf).

    assert (FF: escr_ffun L (b :: acts) (sigma, rr) = true).
    {
      unfold escr_ffun; desf; destruct n; ins.
      repeat eexists; eauto; intros.

      desf.
        eapply interp_escrOK with (sigma := sigma) (r := L (TcondS c) +!+ rE).
        rewrite ESCR; ins; split; eauto; intro X. 
        red in VG; desc; clear - X VG1. 
        rewrite res_plusC, !res_plusA, !(res_plusAC rE), !(res_plusAC (L _)) in VG1.
        by rewrite <- res_plusA, X in VG1.

      eapply CC with (E := (TcondS c :: TescS b :: nil)); ins; desf; ins; eauto.
        by repeat constructor; ins; intro; desf.
      rewrite upds, updo, res_plus_emp_r, (res_plusAC rE), <- res_plusA; ins.
      apply NNPP; intro.
      eapply interp_escrOK with (sigma := sigma) (r := L (TcondS c) +!+ rE).
      by rewrite ESCR; ins; split; eauto; intro X; rewrite X in *.
    }

    assert (M: exists rA_rem, L (TescS a) = rr +!+ rA_rem 
                    /\ res_lt (res_strip rr) rA_rem).
    {
      apply res_add_strip_stengthen.
      destruct VescS as (? & EQ); rewrite <- EQ in *.
      eapply In_split in H0; desf.
      rewrite filter_app, map_app, rsum_app; simpl; desf. 
      ins; rewrite !res_plusA, res_plusAC; vauto.
    }
    desf.

    assert (DEFi: rr +!+ incoming (b :: acts) L b <> Rundef). {
      unfold incoming, in_sb, in_rf, in_esc; intro.
      eapply CC with (E := (TescS a :: map (Tsb^~ b) acts ++
                         map (Trf^~ b) acts ++ 
                         map (Tesc^~ b) acts)). 
        by inv ND; constructor; try rewrite 2!nodup_app; repeat apply conj; try apply nodup_map; auto;
          unfold disjoint; intros until 0; rewrite ?in_app_iff, ?in_map_iff; try congruence;
          try red; intros; desf.


        intros; rewrite ?in_cons_iff, ?in_app_iff, ?in_map_iff in *; ins.
        assert (b0 = b) by (clear IN; desf; ins; desf); clear IN'0; subst.
       
        rewrite mhbE in *; destruct HB; clarify.
          by ins; desf; destruct te; desf; ins; eauto using prefix_closed_hb. 
        eby desf; simpls; eauto using prefix_closed_hb.

      red in CONS; desc.
      clear - ACYC LR INa NIN H1 M; rewrite map_upd_irr; ins.
        rewrite (proj1 LR b b), (proj1 (proj2 LR) b b), 
                (proj1 (proj2 (proj2 LR)) b b), !res_plus_emp_l in H1;
                try (by intro; edestruct ACYC; eauto using hb_in_ct). 

        simpl; rewrite M, !map_app, !rsum_app; simpl. 
        by rewrite !res_plusA in *; rewrite res_plusAC, H1, res_plusC.

      by repeat (rewrite ?in_app_iff, ?in_map_iff; simpl); intro; desf.
    }
    assert (DEF: rr +!+ (rE +!+ r0) +!+ out_esc (b :: acts) L b +!+ L (TcondS b) <> Rundef).
    {
      unfold out_esc; rewrite rsum_map_map_eq_empI, res_plus_emp_r; ins.
      rewrite !res_plusA; intro Z. 
      eapply CC with (E := (TescS a :: TescS b :: TcondS b :: nil)). 
        by repeat constructor; ins; intro; desf.        
        by ins; desf; ins; eauto.
      ins; rewrite upds, !updo; ins; try congruence.
      rewrite res_plus_emp_r, M, !res_plusA.
      by rewrite res_plusAC, !(res_plusAC _ (L (TescS b))), Z, res_plusC.
    }

  assert (MYDEF: (r0 +!+ res_strip rE) +!+ rr <> Rundef).
    rewrite res_plusC, <- res_plusA; apply res_plus_strip_def; rewrite res_plusA.
    by intro X; apply DEF; clear - X; rewrite <- res_plusA, (res_plusC rE); rewrite X.

  eexists (upd (upd (upd L (TescS a) rA_rem) (Tesc a b) (rr +!+ L (Tesc a b)))
               (TcondS b) (L (TcondS b) +!+ rE)), I, ((r0 +!+ res_strip rE) +!+ rr).
  split.
    eapply IN'; split; ins. 
    do 3 eexists; eauto; split; eauto. 
    by intro X; rewrite X in *.

  split.
  ins.
  destruct (eqP a0 a); desf.
  eexists rA, IA; unfold NW; ins.
  unfold incoming, out_esc, out_sb, out_rf, in_sb, in_rf, in_esc in *; ins;
  rupd; try congruence.
  rewrite !map_upd_irr; try (by rewrite in_map_iff; intro; desf).
  repeat (split; try done).
    by revert VG0; rewrite M, !res_plusA, (res_plusAC rr rA_rem).
    by ins; rewrite !updo; try done; eauto.

  assert (A1: L (TcondS b) +!+ rE <> Rundef).
   intro X; destruct DEF; clear - X.
   by rewrite res_plusC, !res_plusA, !(res_plusAC (L (TcondS b))), <- res_plusA, X.

  assert (A2: rr +!+ rA_rem <> Rundef).
    by clear - NN1 M; intro X; rewrite !res_plusA, M, X, !(res_plusAC Rundef) in NN1.

  do 2 (rewrite updC, escr_ffun_upd_irr; ins).
  eapply In_split in H0; desf.
  revert VescS; rewrite M, !filter_app, !map_app, !rsum_app; simpl;
  desf; simpl; rewrite ?(res_plusAC rr);
  rewrite <- !rsum_app, <- !map_app, <- !filter_app.

  unfold escr_ffun in Heq; desf; desf;
  destruct (e1 b (in_eq _ _)); rewrite upds, res_plusC; ins;
  apply RP_ext; try rewrite res_plusC; congruence.

  intro ZZ; try apply res_lt_frame_rev in ZZ; ins.
  eapply helper_escr_ffun; auto.

  clear - DEF NEQ NIN IN V n; ins; specialize (V _ IN); unfold valid_node in *; desf.
  eexists r, I; unfold NW; ins.
  unfold incoming, out_esc, out_sb, out_rf, in_sb, in_rf, in_esc in *; ins;
  rupd; try congruence.
  rewrite !map_upd_irr; try (by rewrite in_map_iff; intro; desf).
  repeat (split; try done).
    by ins; rewrite !updo; try done; eauto.
    eapply helper_escr_ffun; [by rewrite !updo| |by rewrite !escr_ffun_upd_irr].

    by intro X; destruct DEF; rewrite res_plusC, !res_plusA, 
                          !(res_plusAC (L (TcondS b))), <- res_plusA, X.

  split. 
    rupd; try congruence. 
    rewrite updC; ins; eapply compat_upd2_acc; ins; rupd.
    eapply compat_upd3; eauto; ins; rewrite ?M; auto using res_lt_plus_l, res_lt_refl;
      try congruence.
      by rewrite res_plusA, res_plusAC; apply res_lt_refl.
    by rewrite !res_plusA, !(res_plusAC rr), (res_plusC _ rE), res_plus_strip_idemp,
               (res_plusC rE); apply res_lt_refl.

  split.
    by red; unfold out_rf; ins; rewrite !map_upd_irr, !updo in *; ins;
       try (by rewrite in_map_iff; intro; desf); eapply CO.

  inv ND.
  revert VG DEF DEFi.
  rewrite incoming_upd_irr, updC, incoming_upd_irr; try done.
  rewrite incoming_upd_ext_esc; ins; vauto.
  unfold  out_esc, out_rf, in_sb, in_rf; simpl.
  rewrite !map_upd_irr in *; try (by rewrite in_map_iff; intro; desf).
  unfold in_esc; simpl; repeat first [rewrite upds | rewrite updo; [|by intro; desf]].


  repeat (split; auto using in_eq, in_cons); ins; try by intros; rewrite !updo.
  repeat first [rewrite upds | rewrite updo; [|by intro; desf]].
  unfold out_esc in *; ins.
  rewrite (res_plusC _ rE); rewrite !res_plusA in *.
  rewrite !(res_plusAC rE), !(res_plusAC rr).
  rewrite (res_plusAC (res_strip rE)), <- (res_plusA rE), res_plus_strip_idemp.
  red in VG; desc; rewrite VG, VG1 in *. 
  destruct rr; ins; desf;
  do 7 eexists; ins; instantiate; ins; rewrite ?orR, VG2; clear; tauto.


  rewrite res_plusC. 
  rewrite res_lt_strip_l in *; apply res_lt_strip_frame.
  2: by rewrite res_plusC.
  rewrite res_plusC; revert RS; rewrite !res_strip_plus, res_strip_idemp; ins.
  by rewrite res_plusC; intro X; rewrite X in *; desf.

  eapply helper_escr_ffun with rE; rewrite ?updo, ?escr_ffun_upd_irr; ins.
  intro X; destruct DEF; clear - X.
  by rewrite res_plusC, !res_plusA, !(res_plusAC (L (TcondS b))), <- res_plusA, X.

  rupd; congruence.

  SCase "escrow introduced at b".

    assert (exists rr, In (sigma, rr) I).
      destruct (proj1 (vghost_escr VG sigma)); ins.
      apply res_escr_plusI; eauto.
      by intro X; rewrite X in *; red in VG; desf.
    desf.
    assert (ERES: P2 rr) by (eapply VI in H0; desf; congruence).

    assert (FF: escr_ffun L (b :: acts) (sigma, rr) = true).
    {
      unfold escr_ffun; desf; destruct n; ins.
      repeat eexists; eauto; intros.

      eapply CC with (E := (TcondS c :: TescS b :: nil)); try by ins; desf; ins; eauto.
        by repeat constructor; ins; intro; desf.
      clear IN.
      ins.
      rewrite upds, updo, res_plus_emp_r, (res_plusAC rE), <- res_plusA; ins.
      apply NNPP; intro.
      eapply interp_escrOK with (sigma := sigma) (r := L (TcondS c) +!+ rE).
      by rewrite ESCR; ins; split; eauto; intro X; rewrite X in *.
    }

    assert (M: exists rA_rem, L (TescS b) = rr +!+ rA_rem 
                    /\ res_lt (res_strip rr) rA_rem).
    {
      apply res_add_strip_stengthen.
      destruct VES as (? & EQ); rewrite <- EQ in *.
      eapply in_split in H0; desf.
      rewrite filter_app, map_app, rsum_app; simpl; desf. 
      ins; rewrite !res_plusA, res_plusAC; vauto.
    }
    desf.
    
  assert (CDEF: L (TcondS b) +!+ rE <> Rundef).
    intro X; revert VG; rewrite res_plusC, !res_plusA.
    rewrite !(res_plusAC rE), !(res_plusAC (L (TcondS b))), <- res_plusA, X.
    by unfold valid_ghost; ins; desf.

  assert (DDEF: r0 +!+ rr <> Rundef).
    intro X; revert VG; unfold out_esc; rewrite !res_plusA.
    rewrite M, ?res_plusA, !(res_plusAC _ rE), <- res_plusA, X.
    by unfold valid_ghost; ins; desf.

  assert (MYDEF: (r0 +!+ res_strip rE) +!+ rr <> Rundef).
    rewrite res_plusC, <- res_plusA; apply res_plus_strip_def; rewrite res_plusA.
    intro X; red in VG; desc; unfold out_esc in VG0; rewrite M in VG0; clear - X VG0.
    by rewrite <- !res_plusA, (res_plusC rE), (res_plusC _ rr), X in VG0.

  eexists (upd (upd L (TescS b) rA_rem) (TcondS b) (L (TcondS b) +!+ rE)), I, 
          ((r0 +!+ res_strip rE) +!+ rr).
  split. 
    apply IN'; split; ins.
    repeat eexists; eauto.
    by intro X; rewrite X in *. 

  split.
  clear - CDEF NIN V; ins; specialize (V _ IN); unfold valid_node in *; desf.
  eexists r, I; ins.
  unfold incoming, out_esc, out_sb, out_rf, in_sb, in_rf, in_esc in *; ins.
  rewrite !map_upd_irr, !updo; try congruence; try (by rewrite in_map_iff; intro; desf).
  unfold NW; repeat (split; try done).
    by ins; rewrite !updo; eauto.
  by eapply helper_escr_ffun; [by rewrite !updo| |by rewrite !escr_ffun_upd_irr].

  split.
    rewrite updo, upds; ins.
    rewrite updC, updss; ins.
    repeat red; ins; eapply CC; eauto; revert H1.
  destruct (classic (In (TcondS b) E)) as [P|];
    [eapply In_NoDup_Permutation in P; desf;
     rewrite !(Permutation_rsum (Permutation_map _ P)); simpl; try done; rename E into E', l' into E;
     assert (NoDup E) by eauto using nodup_consD, Permutation_nodup |];
    rewrite map_upd_irr, ?upds, ?updo; eauto; simpl; try done;
  (destruct (classic (In (TescS b) E)) as [P'|];
    [eapply In_NoDup_Permutation in P'; desf;
     rewrite !(Permutation_rsum (Permutation_map _ P')); simpl; try done; rename E into E'', l' into E;
     assert (NoDup E) by eauto using nodup_consD, Permutation_nodup |];
    rewrite !map_upd_irr, ?upds; eauto; simpl; try done). 

  intro X; rewrite M; rewrite <- X; clear; rewrite !res_plusA.
  rewrite 2!(res_plusAC (res_strip rE)), <- (res_plusA _ (res_strip rE)), res_plus_strip_idemp.
  by f_equal; rewrite !(res_plusAC rr); f_equal; rewrite !(res_plusAC rE).

  intro X; exfalso.
  assert (~ In (TescS b) E') by (intro Y; apply (Permutation_in _ P) in Y; ins; desf).
  apply CC with (E := TescS b :: E'); ins; desf; eauto; vauto.
    by eapply (IND (TcondS b)) in HB; eauto using Permutation_in, Permutation_sym, in_eq.
  rewrite upds, map_upd_irr; try done.
  rewrite !(Permutation_rsum (Permutation_map _ P)); simpl. 
  clear - X; rewrite !res_plusA in *.
  by rewrite (res_plusAC r0), (res_plusAC _ rE), X, <- res_plusA, res_plusC.

  intro X; rewrite M; clear - X; rewrite !res_plusA, !(res_plusAC rr) in *. 
  rewrite <- (res_plus_strip_idemp rE), !res_plusA, (res_plusAC r0).
  by rewrite !(res_plusAC rE), X, res_plusC.

  split.
    red; ins; unfold out_rf; ins.
    rewrite !map_upd_irr, !updo; try congruence; try (by rewrite in_map_iff; intro; desf).
    by apply CO.

  revert VG.
  rewrite !incoming_upd_irr; try done.
  unfold out_esc, out_rf, in_esc, in_sb, in_rf; simpl.
  rewrite !map_upd_irr in *; try (by rewrite in_map_iff; intro; desf).
  unfold in_esc; simpl; repeat first [rewrite upds | rewrite updo; [|by intro; desf]].

  repeat (split; auto using in_eq, in_cons); ins; try by intros; rewrite !updo.
    revert VG; rewrite M.
    rewrite (res_plusC _ rE); rewrite !res_plusA in *.
    rewrite !(res_plusAC rE), !(res_plusAC rr).
    by rewrite (res_plusAC (res_strip rE)), <- (res_plusA _ (res_strip rE)),
               res_plus_strip_idemp.

  rewrite res_plusC; apply res_lt_plus_l.
  rewrite res_lt_strip_l, res_strip_plus, res_strip_idemp in *; try done; 
    try by intro X; rewrite X in *.
  by rewrite res_plusC.

  revert VES; rewrite M.
  rewrite updC, escr_ffun_upd_irr; try done.

  eapply In_split in H0; desf.
  rewrite !filter_app, !map_app, !rsum_app; simpl;
  desf; simpl; rewrite ?(res_plusAC rr);
  rewrite <- !rsum_app, <- !map_app, <- !filter_app.

  unfold escr_ffun in Heq; desf; desf;
  destruct (e1 b (in_eq _ _)); rewrite upds, res_plusC; ins;
  apply RP_ext; try rewrite res_plusC; congruence.

  intro ZZ; try apply res_lt_frame_rev in ZZ; ins.
  eapply helper_escr_ffun; auto.
  rewrite <- M; intro X; red in VG; desf; clear - X VG0.
  by rewrite X in *; ins; rewrite res_plusC in VG0.
Grab Existential Variables. auto.
Qed.

